#cython: infer_types=True
#cython: cdivision=True
from libc.math cimport sqrt, M_PI, log, pow, abs
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef (double, int) predict_cauchy(double[::1] next_vec, double[::1] prev_vec, float alpha, int lag, float precision, int dimension):
    cdef int delta_t, abnormal
    cdef double log_pdf
    log_pdf = 0.0
    abnormal = 0
    delta_t = lag + 1

    for dim_i in range(dimension):
        vec1 = next_vec[dim_i]
        vec2 = prev_vec[dim_i]
        if vec1 < 0:
            vec1 = vec1 - precision
        else:
            vec1 = vec1 + precision
        if vec2 < 0:
            vec2 = vec2 - precision
        else:
            vec2 = vec2 + precision
        coord_ratio = vec1 / vec2

        if 0.95 < alpha < 1.05:
            if abs(coord_ratio) > 8: ## TODO
                abnormal = 1
            log_pdf += log( 1/M_PI * 1/((coord_ratio)*(coord_ratio) + 1) )

        else:
            rho = 1/2. * (pow((delta_t-1), alpha) - 2*pow(delta_t, alpha) + pow((delta_t+1), alpha))
            relativ_cov = 1/2. * (pow((delta_t+1), alpha) - pow((delta_t), alpha) - pow((1), alpha))
            scale = sqrt(abs(1-(rho*rho))) ## TODO
            if abs(coord_ratio-rho) > 8 * scale: ## TODO
                abnormal = 1
            log_pdf += log( 1/(M_PI * scale) * 1 / ( ((coord_ratio - relativ_cov)/scale)*((coord_ratio - relativ_cov)/scale) * (rho/relativ_cov) + (relativ_cov/rho) ) )

    return log_pdf, abnormal


#cython: infer_types=True
#cython: cdivision=True
from libc.math cimport sqrt, M_PI, log
import numpy as np
cimport cython
from scipy.linalg import lstsq as sp_lstsq
from scipy.linalg import solve


def ord_lu(X, y):
    A = X.T @ X
    b = X.T @ y
    beta = solve(A, b, overwrite_a=True, overwrite_b=True,
                 check_finite=False)
    return beta
    

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef pack_vars(float[::1] vars, int len_img):
    cdef float a,b,c,d,e,f
    cdef Py_ssize_t i
    var_stack = np.zeros([len_img, 6], dtype=np.float32)

    a = -1. / (2 * vars[0] * (1 - vars[4]*vars[4]))
    b = vars[1] / ((1 - vars[4] * vars[4]) * vars[0]) - (vars[4] * vars[3]) / ((1 - vars[4] * vars[4]) * sqrt(vars[0]) * sqrt(vars[2]))
    c = -1. / (2 * vars[2] * (1 - vars[4]*vars[4]))
    d = vars[3] / ((1 - vars[4] * vars[4]) * vars[2]) - (vars[4] * vars[1]) / ((1 - vars[4] * vars[4]) * sqrt(vars[0]) * sqrt(vars[2]))
    e = vars[4] / ((1 - vars[4] * vars[4]) * sqrt(vars[0]) * sqrt(vars[2]))
    f = (-(vars[1] * vars[1])/(2 * (1 - vars[4] * vars[4]) * vars[0]) - (vars[3] * vars[3])/(2 * (1 - vars[4] * vars[4]) * vars[2]) + (vars[4] * vars[1] * vars[3]) / ((1 - vars[4] * vars[4]) * sqrt(vars[0]) * sqrt(vars[2])) + 
        log(1 / (2 * M_PI * sqrt(vars[0]) * sqrt(vars[2]) * (sqrt(1 - vars[4] * vars[4])))))

    for i in range(len_img):
        var_stack[i][0] = a
        var_stack[i][1] = b
        var_stack[i][2] = c
        var_stack[i][3] = d
        var_stack[i][4] = e
        var_stack[i][5] = f
    return var_stack


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef matrix_decomp(matrix, q):
    ret_mat = []
    for x in range(0, len(matrix), q):
        ret_mat.append(matrix[x: min(x+q, len(matrix))])
    return ret_mat


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef unpack_coefs(coefs, window_size):
    err_indices = []
    x_mu = []
    y_mu = []
    rho = coefs[:, 4] * np.sqrt(1/(4 * -abs(coefs[:, 0]) * -abs(coefs[:, 2])))
    k = 1 - rho**2
    x_var = abs(1/(-2 * coefs[:, 0] * k))
    y_var = abs(1/(-2 * coefs[:, 2] * k))
    for err_indice, (xvar_check, yvar_check, r) in enumerate(zip(x_var, y_var, rho)):
        if xvar_check < 0 or yvar_check < 0 or xvar_check > 3 * window_size[0] or yvar_check > 3 * window_size[1] or r < -1 or r > 1 or np.isnan(r):
            #print(r, np.isnan(r))
            err_indices.append(err_indice)

    for i, (b, d) in enumerate(zip(coefs[:, 1], coefs[:, 3])):
        if i in err_indices:
            x_mu.extend([0])
            y_mu.extend([0])
        else:
            coef_mat = np.array([[-rho[i] * np.sqrt(y_var[i]) / np.sqrt(x_var[i]), 1.],
                                 [1., -rho[i] * np.sqrt(x_var[i]) / np.sqrt(y_var[i])]])
            ans_mat = np.array([[d * k[i] * y_var[i]], [b * k[i] * x_var[i]]])
            x_, y_, = np.linalg.lstsq(coef_mat, ans_mat, rcond=None)[0]
            x_mu.extend(x_)
            y_mu.extend(y_)
    x_mu = np.array(x_mu)
    y_mu = np.array(y_mu)
    amp = np.exp(coefs[:, 5] + (x_mu**2 / (2 * k * x_var)) + (y_mu**2 / (2 * k * y_var)) - (rho * x_mu * y_mu / (k * np.sqrt(x_var) * np.sqrt(y_var))))
    return [x_var, x_mu, y_var, y_mu, rho, amp], err_indices


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef guo_algorithm(imgs:np.ndarray, bgs, float[::1] p0, 
                    xgrid, ygrid, window_size=(7, 7), repeat=7, decomp_n=1):
    cdef int k, nb_imgs
    cdef float [:,::1] imgs_view, bgs_view
    cdef float [::1] local_backgrounds

    k = 0
    nb_imgs = imgs.shape[0]
    coef_vals = pack_vars(p0, nb_imgs)
    img_view = imgs
    bgs_view = bgs
    ## background for each crop image needed rather than background intensity for whole image.
    #imgs = np.maximum(np.zeros(imgs.shape), imgs - bgs.reshape(-1, window_size[0], window_size[1])) + 1e-2
    local_backgrounds = element_wise_subtraction_2d(img_view, bgs_view)
    element_wise_maximum_2d(img_view, local_backgrounds)
    yk_2 = imgs.copy()

    while k < repeat:
        if k != 0:
            yk_2 = np.exp(coef_vals[:, 0].reshape(-1, 1) * xgrid**2 + coef_vals[:, 1].reshape(-1, 1) * xgrid +
                          coef_vals[:, 2].reshape(-1, 1) * ygrid**2 + coef_vals[:, 3].reshape(-1, 1) * ygrid +
                          coef_vals[:, 4].reshape(-1, 1) * xgrid * ygrid + coef_vals[:, 5].reshape(-1, 1))
        yk_2 = yk_2 * yk_2
        coef1 = yk_2 * xgrid * xgrid * xgrid * xgrid
        coef2 = yk_2 * xgrid * xgrid * xgrid
        coef3 = yk_2 * xgrid**2 * ygrid**2
        coef4 = yk_2 * xgrid**2 * ygrid
        coef5 = yk_2 * xgrid * xgrid * xgrid * ygrid
        coef6 = yk_2 * xgrid**2
        coef7 = yk_2 * xgrid * ygrid**2
        coef8 = yk_2 * xgrid * ygrid
        coef9 = yk_2 * xgrid
        coef10 = yk_2 * ygrid * ygrid * ygrid * ygrid
        coef11 = yk_2 * ygrid * ygrid * ygrid
        coef12 = yk_2 * xgrid * ygrid * ygrid * ygrid
        coef13 = yk_2 * ygrid**2
        coef14 = yk_2 * ygrid
        coef15 = yk_2
        coef_matrix = np.sum(
            np.array(
                [[coef1, coef2, coef3, coef4, coef5, coef6],
                 [coef2, coef6, coef7, coef8, coef4, coef9],
                 [coef3, coef7, coef10, coef11, coef12, coef13],
                 [coef4, coef8, coef11, coef13, coef7, coef14],
                 [coef5, coef4, coef12, coef7, coef3, coef8],
                 [coef6, coef9, coef13, coef14, coef8, coef15]]
            ), axis=(3)).transpose(2, 0, 1)
        ans1 = xgrid ** 2 * yk_2 * np.log(imgs)
        ans2 = xgrid * yk_2 * np.log(imgs)
        ans3 = ygrid ** 2 * yk_2 * np.log(imgs)
        ans4 = ygrid * yk_2 * np.log(imgs)
        ans5 = xgrid * ygrid * yk_2 * np.log(imgs)
        ans6 = yk_2 * np.log(imgs)
        ans_matrix = np.sum(
            np.array(
                [[ans1], [ans2], [ans3], [ans4], [ans5], [ans6]]
            ), axis=(3), dtype=np.float32).transpose(2, 0, 1)
        coef_matrix = matrix_decomp(coef_matrix, decomp_n)
        ans_matrix = matrix_decomp(ans_matrix, decomp_n)
        decomp_coef_vals = matrix_decomp(coef_vals, decomp_n)
        x_matrix = []
        for (a_mats, b_mats, coef_val) in zip(coef_matrix, ans_matrix, decomp_coef_vals):
            a_mat = np.zeros((a_mats.shape[0] * a_mats.shape[1], a_mats.shape[0] * a_mats.shape[2]))
            for x, vals in zip(range(0, a_mat.shape[0], 6), a_mats):
                a_mat[x:x+6, x:x+6] = vals
            b_mat = b_mats.flatten().reshape(-1, 1)
            x_matrix.extend(sp_lstsq(a_mat, b_mat, lapack_driver='gelsy', check_finite=False)[0])
        x_matrix = np.array(x_matrix).reshape(-1, 6)
        if np.allclose(coef_vals, x_matrix, rtol=1e-7):
            break
        coef_vals = x_matrix
        k += 1
    return coef_vals


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef element_wise_maximum_2d(float [:,::1] array1, float[:] local_bgs):
    cdef int nb_imgs, img_size
    cdef Py_ssize_t i, j
    nb_imgs = array1.shape[0]
    img_size = array1.shape[1]
    
    for i in range(nb_imgs):
        for j in range(img_size):
            array1[i][j] = array1[i][j] - local_bgs[i] + 1e-2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef element_wise_subtraction_2d(float [:,::1] array1, float [:,::1] array2):
    assert(array1.shape[0] == array2.shape[0] and array1.shape[1] == array2.shape[1])
    cdef int nb_imgs, img_size
    cdef float local_bg
    cdef Py_ssize_t i, j
    nb_imgs = array1.shape[0]
    img_size = array1.shape[1]
    local_bgs = np.empty(nb_imgs, dtype=np.float32)
    cdef float[:] local_bgs_view = local_bgs

    for i in range(nb_imgs):
        local_bg = 9999.0
        for j in range(img_size):
            array1[i][j] = array1[i][j] - array2[i][j]
            local_bg = min(local_bg, array1[i][j])
        local_bgs_view[i] = local_bg
    return local_bgs


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef matrix_pow_2d(float [:,::1] array1, int power):
    cdef int row_size, col_size
    cdef Py_ssize_t i, j, k
    cdef float tmp
    row_size = array1.shape[0]
    col_size = array1.shape[1]
    assert(power >= 0)
    if power == 0:
        for i in range(row_size):
            for j in range(col_size):
                array1[i][j] = 1.0 
    elif power >= 2:
        power -= 1
        for i in range(row_size):
            for j in range(col_size):
                    tmp = array1[i][j]
                    for k in range(power):
                        array1[i][j] = array1[i][j] * tmp



import numpy as np
import pandas as pd
import networkx as nx
from itertools import product, permutations
from FreeTrace.module import data_load, data_save
from tqdm import tqdm
import math
import cv2
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib as mpl
import matplotlib.pyplot as plt



def simple_preprocessing(data, pixelmicrons, framerate, cutoff=[3, 99999], selected_state=0, div_time_gap=True, tamsd_calcul=True, color_palette = ['red','cyan','green','blue','gray','pink']):
    # load FreeTrace+Bi-ADD data without NaN (NaN where trajectory length is shorter than 5, default in BI-ADD)
    assert 'yellow' not in color_palette, "Yellow can't be included in the color palette, please exclude yellow."
    data = data.dropna()
    # using dictionary to convert specific columns
    convert_dict = {'state': int}
    data = data.astype(convert_dict)
    traj_indices = pd.unique(data['traj_idx'])
    total_states = sorted(data['state'].unique())

    if len(data) == 0:
        print("*** Data is empty. ***")
        return
    
    if selected_state not in total_states:
        print("*** Selected state is not found in the data. Please try other state numbers. ***")
        return
    
    if len(total_states) == 1:
        color_palette = ['red']

    # state re-ordering w.r.t. K
    avg_ks = []
    for st in total_states:
        avg_ks.append(data['K'][data['state']==st].mean())
    avg_ks = np.array(avg_ks)
    prev_states = np.argsort(avg_ks)
    state_reorder = {st:idx for idx, st in enumerate(prev_states)}
    ordered_states = np.empty(len(data['state']), dtype=np.uint8)
    for st_idx in range(len(ordered_states)):
        ordered_states[st_idx] = state_reorder[data['state'].iloc[st_idx]]
    data['state'] = ordered_states

    # initializations
    dim = 2 # will be changed in future.
    max_frame = data.frame.max()

    product_states = list(product(total_states, repeat=2))
    state_graph = nx.DiGraph()
    state_graph.add_nodes_from(total_states)
    state_graph.add_edges_from(product_states, count=0, freq=0)
    
    if len(total_states) > 1:
        state_changing_duration = {tuple(state): [] for state in product_states}
    else:
        state_changing_duration = {tuple([total_states[0], total_states[0]]):[]}

    state_markov = [[0 for _ in range(len(total_states))] for _ in range(len(total_states))]
    analysis_data1 = {}
    analysis_data1[f'mean_jump_d'] = []
    analysis_data1[f'K'] = []
    analysis_data1[f'H'] = []
    analysis_data1[f'state'] = []
    analysis_data1[f'duration'] = []
    analysis_data1[f'traj_id'] = []
    analysis_data1[f'color'] = []

    analysis_data2 = {}
    analysis_data2[f'2d_displacement'] = []
    analysis_data2[f'state'] = []

    analysis_data3 = {}
    analysis_data3[f'angle'] = []
    analysis_data3[f'state'] = []

    analysis_data4 = {}
    for st in total_states:
        analysis_data4[st] = [[] for _ in range(9999)]
        analysis_data4[st][0].append(0)

    analysis_data5 = {}
    analysis_data5[f'1d_ratio'] = []
    analysis_data5[f'state'] = []

    msd_ragged_ens_trajs = {st:[] for st in total_states}
    tamsd_ragged_ens_trajs = {st:[] for st in total_states}
    msd = {}
    msd[f'mean'] = []
    msd[f'std'] = []
    msd[f'nb_data'] = []
    msd[f'state'] = []
    msd[f'time'] = []
    tamsd = {}
    tamsd[f'mean'] = []
    tamsd[f'std'] = []
    tamsd[f'nb_data'] = []
    tamsd[f'state'] = []
    tamsd[f'time'] = []

    # get data from trajectories
    if tamsd_calcul:
        print("Computing TAMSD... This takes while depending on the length and number of trajectories.")
        
    for traj_idx in tqdm(traj_indices, ncols=120, desc=f'Analysis', unit=f'trajectory'):
        single_traj = data.loc[data['traj_idx'] == traj_idx].copy()
        single_traj = single_traj.sort_values(by=['frame'])

        # chunk into sub-trajectories by frame gap
        before_st = single_traj.state.iloc[0]
        frame_gaps = np.diff(single_traj.frame.to_numpy())
        time_chunk_idx = [0, len(single_traj)]
        if div_time_gap:
            for f_idx, frame_gap in enumerate(frame_gaps):
                if frame_gap > 1.0:
                    time_chunk_idx.append(f_idx+1)
            time_chunk_idx = sorted(time_chunk_idx)

        for i in range(len(time_chunk_idx) - 1):
            time_chunked_single_traj = single_traj.iloc[time_chunk_idx[i]:time_chunk_idx[i+1]].copy()

            # chunk into sub-trajectories by state
            before_st = time_chunked_single_traj.state.iloc[0]
            state_chunk_idx = [0, len(time_chunked_single_traj)]
            for st_idx, st in enumerate(time_chunked_single_traj.state):
                if st != before_st:
                    state_chunk_idx.append(st_idx)
                before_st = st
            state_chunk_idx = sorted(state_chunk_idx)

            for i in range(len(state_chunk_idx) - 1):
                sub_trajectory = time_chunked_single_traj.iloc[state_chunk_idx[i]:state_chunk_idx[i+1]].copy()
                # trajectory length filter condition
                if cutoff[0] <= len(sub_trajectory) <= cutoff[1]:
                    # state of trajectory
                    state = sub_trajectory.state.iloc[0]
                    fr_H = sub_trajectory.H.iloc[0]
                    fr_K = sub_trajectory.K.iloc[0]

                    # convert from pixel-coordinate to micron.
                    sub_trajectory.x *= pixelmicrons
                    sub_trajectory.y *= pixelmicrons
                    sub_trajectory.z *= pixelmicrons 
                    #fr_K *= ((pixelmicrons**2)) #/ (framerate**fr_H))

                    frame_diffs = sub_trajectory.frame.iloc[1:].to_numpy() - sub_trajectory.frame.iloc[:-1].to_numpy()
                    duration = np.sum(frame_diffs) * framerate

                    if len(state_chunk_idx) == 2:
                        state_graph[state][state]['count'] += 1
                    else:
                        if i >= 1:
                            state_graph[prev_state][state]['count'] += 1
                            state_changing_duration[tuple([prev_state, state])].append(prev_duration)
                        if i == len(state_chunk_idx) - 2:  # last state of chunked trjectory which is considered as non-state changing stae.
                            state_graph[state][state]['count'] += 1 
                    prev_state = state
                    prev_duration = duration
                
                    # coordinate rescale
                    sub_trajectory.x -= sub_trajectory.x.iloc[0]
                    sub_trajectory.y -= sub_trajectory.y.iloc[0]
                    sub_trajectory.z -= sub_trajectory.z.iloc[0]

                    # calcultae jump distances (assumption of Brownian particle if the trajectory is not divided by the gap in frames)               
                    jump_distances = (np.sqrt(((sub_trajectory.x.iloc[1:].to_numpy() - sub_trajectory.x.iloc[:-1].to_numpy()) ** 2) / (sub_trajectory.frame.iloc[1:].to_numpy() - sub_trajectory.frame.iloc[:-1].to_numpy())
                                            + ((sub_trajectory.y.iloc[1:].to_numpy() - sub_trajectory.y.iloc[:-1].to_numpy()) ** 2) / (sub_trajectory.frame.iloc[1:].to_numpy() - sub_trajectory.frame.iloc[:-1].to_numpy()) )) 
                
                    # angles
                    x_vec = (sub_trajectory.x.iloc[1:].to_numpy() - sub_trajectory.x.iloc[:-1].to_numpy())
                    y_vec = (sub_trajectory.y.iloc[1:].to_numpy() - sub_trajectory.y.iloc[:-1].to_numpy())
                    vecs = np.vstack([x_vec, y_vec]).T
                    angles = [dot_product_angle(vecs[vec_idx], vecs[vec_idx+1]) for vec_idx in range(len(vecs) - 1)]
                    ratios = np.array([])
                    for l in range(1, 2):
                        r_tmp = np.concatenate([(x_vec[l::l] / x_vec[:-l:l]), (y_vec[l::l] / y_vec[:-l:l])])
                        ratios = np.concatenate((ratios, r_tmp))
                    #ratios = [0 if math.isnan(val) or math.isinf(val) else val for val in ratios]

                    # Ensemble averaged SD
                    copy_frames = sub_trajectory.frame.to_numpy()
                    copy_frames = copy_frames - copy_frames[0]
                    tmp_msd = []
                    for frame, sq_disp in zip(np.arange(0, copy_frames[-1], 1), ((sub_trajectory.x.to_numpy())**2 + (sub_trajectory.y.to_numpy())**2) / dim / 2):
                        if frame in copy_frames:
                            tmp_msd.append(sq_disp)
                        else:
                            tmp_msd.append(None)
                    msd_ragged_ens_trajs[state].append(tmp_msd)


                    # TAMSD
                    if tamsd_calcul:
                        tamsd_tmp = []
                        for lag in range(len(sub_trajectory)):
                            if lag == 0:
                                tamsd_tmp.append(0)
                            else:
                                time_averaged = []
                                for pivot in range(len(sub_trajectory) - lag):
                                    if lag == (sub_trajectory.frame.iloc[pivot + lag] - sub_trajectory.frame.iloc[pivot]):
                                        time_averaged.append(((sub_trajectory.x.iloc[pivot + lag] - sub_trajectory.x.iloc[pivot]) ** 2 + (sub_trajectory.y.iloc[pivot + lag] - sub_trajectory.y.iloc[pivot]) ** 2) / dim / 2)
                                        analysis_data4[state][lag].append((sub_trajectory.x.iloc[pivot + lag] - sub_trajectory.x.iloc[pivot]))
                                        analysis_data4[state][lag].append((sub_trajectory.y.iloc[pivot + lag] - sub_trajectory.y.iloc[pivot]))
                                if len(time_averaged) > 0:
                                    tamsd_tmp.append(np.mean(time_averaged))
                                else:
                                    tamsd_tmp.append(None)
                    else:
                        tamsd_tmp = [0] * len(sub_trajectory)
                    tamsd_ragged_ens_trajs[state].append(tamsd_tmp)


                    if len(state_chunk_idx) > 2:
                        analysis_data1[f'color'].append("yellow")
                    else:
                        analysis_data1[f'color'].append(color_palette[state])

                    # add data1 for the visualization
                    analysis_data1[f'mean_jump_d'].append(jump_distances.mean())
                    analysis_data1[f'K'].append(fr_K)
                    analysis_data1[f'H'].append(fr_H)
                    analysis_data1[f'state'].append(state)
                    analysis_data1[f'duration'].append(duration)
                    analysis_data1[f'traj_id'].append(sub_trajectory.traj_idx.iloc[0])


                    # add data2 for the visualization
                    analysis_data2[f'2d_displacement'].extend(list(jump_distances))
                    analysis_data2[f'state'].extend([sub_trajectory.state.iloc[0]] * len(list(jump_distances)))

                    # add data3 for angles
                    analysis_data3[f'angle'].extend(list(angles))
                    analysis_data3[f'state'].extend([sub_trajectory.state.iloc[0]] * len(list(angles)))

                    # add data5 for ratio (consecutive displacements)
                    analysis_data5[f'1d_ratio'].extend(list(ratios))
                    analysis_data5[f'state'].extend([sub_trajectory.state.iloc[0]] * len(list(ratios)))



    # calculate average of msd and tamsd for each state
    for state_key in total_states:
        msd_mean = []
        msd_std = []
        msd_nb_data = []
        tamsd_mean = []
        tamsd_std = []
        tamsd_nb_data = []
        for t in range(max_frame):
            msd_nb_ = 0
            tamsd_nb_ = 0
            msd_row_data = []
            tamsd_row_data = []
            for row in range(len(msd_ragged_ens_trajs[state_key])):
                if t < len(msd_ragged_ens_trajs[state_key][row]) and msd_ragged_ens_trajs[state_key][row][t] is not None:
                    msd_row_data.append(msd_ragged_ens_trajs[state_key][row][t])
                    msd_nb_ += 1
            for row in range(len(tamsd_ragged_ens_trajs[state_key])):
                if t < len(tamsd_ragged_ens_trajs[state_key][row]) and tamsd_ragged_ens_trajs[state_key][row][t] is not None:
                    tamsd_row_data.append(tamsd_ragged_ens_trajs[state_key][row][t])
                    tamsd_nb_ += 1
            msd_mean.append(np.mean(msd_row_data))
            msd_std.append(np.std(msd_row_data))
            msd_nb_data.append(msd_nb_)
            tamsd_mean.append(np.mean(tamsd_row_data))
            tamsd_std.append(np.std(tamsd_row_data))
            tamsd_nb_data.append(tamsd_nb_)

        sts = [state_key] * max_frame
        times = np.arange(0, max_frame) * framerate

        msd[f'mean'].extend(msd_mean)
        msd[f'std'].extend(msd_std)
        msd[f'nb_data'].extend(msd_nb_data)
        msd[f'state'].extend(sts)
        msd[f'time'].extend(times)
        tamsd[f'mean'].extend(tamsd_mean)
        tamsd[f'std'].extend(tamsd_std)
        tamsd[f'nb_data'].extend(tamsd_nb_data)
        tamsd[f'state'].extend(sts)
        tamsd[f'time'].extend(times)

    # normalize markov chain
    for edge in state_graph.edges:
        src, dest = edge
        weight = state_graph[src][dest]["count"]
        state_markov[src][dest] = weight
    state_markov = np.array(state_markov, dtype=np.float64)
    for idx in range(len(total_states)):
        state_markov[idx] /= np.sum(state_markov[idx])
    for edge in state_graph.edges:
        src, dest = edge
        state_graph[src][dest]['freq'] = np.round(state_markov[src][dest], 3)


    analysis_data1 = pd.DataFrame(analysis_data1).astype({'state': int, 'duration': float, 'traj_id':str})
    analysis_data2 = pd.DataFrame(analysis_data2)
    analysis_data3 = pd.DataFrame(analysis_data3)
    analysis_data5 = pd.DataFrame(analysis_data5)
    msd = pd.DataFrame(msd)
    tamsd = pd.DataFrame(tamsd)

    if tamsd_calcul == False:
        tamsd = None
        
    print('** preprocessing finished **')
    return analysis_data1, analysis_data2, analysis_data3, analysis_data4, analysis_data5, state_markov, state_graph, msd, tamsd, total_states, state_changing_duration


def count_cumul_trajs_with_roi(data:pd.DataFrame|str, roi_file:str|None, start_frame=1, end_frame=100, cutoff=5):
    from roifile import ImagejRoi
    """
    Cropping trajectory result with ROI(region of interest) or frames.
    It returns a list which contains the number of accumulated trajectories, only considering the first positions and time for each trajectory.

    data: .h5 or equivalent format of trajectory result file.
    roi_file: region_of_interest.roi which containes the roi information in pixel.
    start_frame: start frame to crop the trajectory result with frame.
    end_frame: end frame to crop the trajectory result with frame.
    """
    assert "_traces.csv" in data or type(data) is pd.DataFrame or ".h5" in data, "The input filename should be .h5 or traces.csv which are the results of FreeTrace or BI-ADD or equilvalent format."
    assert end_frame > start_frame, "The number of end frame must be greater than start frame."
    

    if roi_file is None:
        contours = None
    elif type(roi_file) is str and len(roi_file) == 0:
        contours = None
    else:
        contours = ImagejRoi.fromfile(roi_file).coordinates().astype(np.int32)
    
    trajectory_counts = []
    cumul=1
    coords = {t:[] for t in np.arange(start_frame, end_frame+1)}
    time_steps = np.arange(start_frame, end_frame, 1)
    observed_max_t = -1
    observed_min_t = 99999999

    if type(data) is str and '.h5' in data:
        data = data_load.read_multiple_h5s(path=data)
        traj_indices = data['traj_idx'].unique()
        for traj_idx in traj_indices:
            single_traj = data[data['traj_idx'] == traj_idx]
            # considers only the first position.
            x = single_traj['x'].iloc[0]
            y = single_traj['y'].iloc[0]
            t = single_traj['frame'].iloc[0]
            if t in coords and len(single_traj) >= cutoff:
                coords[t].append([x, y])
            observed_max_t = max(t, observed_max_t)
            observed_min_t = min(t, observed_min_t)

    elif type(data) is str and '_traces.csv' in data:
        trajectory_list = data_load.read_trajectory(data)
        for trajectory in trajectory_list:
            xyz = trajectory.get_positions()
            x = xyz[:, 0][0]
            y = xyz[:, 1][0]
            t = int(trajectory.get_times()[0])
            if t in coords and len(xyz) >= cutoff:
                coords[t].append([x, y])
            observed_max_t = max(t, observed_max_t)
            observed_min_t = min(t, observed_min_t)
    else:
        traj_indices = data['traj_idx'].unique()
        for traj_idx in traj_indices:
            single_traj = data[data['traj_idx'] == traj_idx]
            # considers only the first position.
            x = single_traj['x'].iloc[0]
            y = single_traj['y'].iloc[0]
            t = single_traj['frame'].iloc[0]
            if t in coords and len(single_traj) >= cutoff:
                coords[t].append([x, y])
            observed_max_t = max(t, observed_max_t)
            observed_min_t = min(t, observed_min_t)
                

    for t in tqdm(time_steps, ncols=120, desc=f'Accumulation', unit=f'frame'):
        if t == start_frame:
            st_tmp = []
            for stack_t in range(t, t+cumul):
                time_st = []
                if stack_t in time_steps:
                    for stack_coord in coords[stack_t]:
                        if roi_file is not None:
                            x, y = stack_coord
                            masked = cv2.pointPolygonTest(contours, (x, y), False)
                            if masked == 1:
                                time_st.append(stack_coord)
                        else:
                            time_st.append(stack_coord)
                st_tmp.append(time_st)
            traj_count = np.sum([len(x) for x in st_tmp])
            trajectory_counts.append(traj_count)
            prev_tmps=st_tmp
        else:
            stack_t = t+cumul-1
            time_st = []
            if stack_t in time_steps:
                for stack_coord in coords[stack_t]:

                    if roi_file is not None:
                        x, y = stack_coord
                        masked = cv2.pointPolygonTest(contours, (x, y), False)
                        if masked == 1:
                            time_st.append(stack_coord)
                    else:
                        time_st.append(stack_coord)

            st_tmp = prev_tmps[1:]
            st_tmp.append(time_st)
            traj_count = np.sum([len(x) for x in st_tmp])
            trajectory_counts.append(traj_count)
            prev_tmps = st_tmp

    return trajectory_counts, np.cumsum(trajectory_counts)


def check_roi_passing_traces(h5_file:str, roi_file:str|None):
    assert ".h5" in h5_file, "Wrong trajectory file format, .h5 extendsion is needed."
    assert roi_file is not None, "This needs ROI file."

    from roifile import ImagejRoi
    contours = ImagejRoi.fromfile(roi_file).coordinates().astype(np.int32)

    stay_inside_trajectories = []
    out_to_in_trajectories = []
    in_to_out_trajectories = []
    complex_trajectories = []
    df = data_load.read_h5(h5_file)[0]
    traj_indices = df['traj_idx'].unique()

    for traj_idx in traj_indices:
        single_traj = df[df['traj_idx'] == traj_idx]
        xs = single_traj['x'].to_numpy()
        ys = single_traj['y'].to_numpy()
        masks = []

        for x, y in zip(xs, ys):
            masked = int(cv2.pointPolygonTest(contours, (x, y), False))
            masks.append(masked)
        
        if -1 in masks:
            if masks[0] == -1 and masks[-1] == 1:
                out_to_in_trajectories.append(traj_idx)
            elif masks[0] == 1 and masks[-1] == -1:
                in_to_out_trajectories.append(traj_idx)
            else:
                complex_trajectories.append(traj_idx)
        else:
            stay_inside_trajectories.append(traj_idx)
        
    print(f"Filename: {h5_file}\nROIname: {roi_file}")
    print(f'Nb of total trajectories: {len(traj_indices)}')
    print(f'Nb of Out-to-In trajectories: {len(out_to_in_trajectories)},  ratio: {np.round(len(out_to_in_trajectories) / len(traj_indices),3)}')
    print(f'Nb of In-to-Out trajectories: {len(in_to_out_trajectories)},  ratio: {np.round(len(in_to_out_trajectories) / len(traj_indices),3)}')
    print(f'Nb of Complex trajectories: {len(complex_trajectories)},  ratio: {np.round(len(complex_trajectories) / len(traj_indices),3)}')
    print(f'')


def linear_fit(msd:pd.DataFrame, timepoints:dict, states:list):
    assert len(timepoints) == len(states), "The number of timepoints and states should be same."
    slopes = {state:[None, None] for state in states}
    for state in states:
        timepoint = timepoints[state]
        times = msd[msd['state']==state]['time'].to_numpy()
        means = msd[msd['state']==state]['mean'].to_numpy()
        selected_times = times[times <= timepoint]
        selected_means = means[times <= timepoint]
        slope, b = np.polyfit(selected_times, selected_means, 1)
        slopes[state] = [slope, b]
    return slopes


def diffusion_coefficient(msd:pd.DataFrame, timepoints:dict, states:list):
    assert len(timepoints) == len(states), "The number of timepoints and states should be same."
    diff_coef_state = {state:None for state in states}
    for state in states:
        timepoint = timepoints[state]
        diff_coef = []
        times = msd[msd['state']==state]['time'].to_numpy()
        means = msd[msd['state']==state]['mean'].to_numpy()
        selected_times = times[times <= timepoint]
        selected_means = means[times <= timepoint]
        for lag in range(1, len(selected_means)):
            time_lag = selected_times[lag]
            diff_coef_lag = []
            for pivot in range(0, len(selected_means) - lag):
                diff_coef_lag.append(abs(selected_means[lag+pivot] - selected_means[pivot]))
            diff_coef.append(np.mean(diff_coef_lag) / time_lag)
        diff_coef_state[state] = np.round(np.mean(diff_coef), 4)
    return diff_coef_state


def unit_vector(vector):
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.degrees(np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)))


def dot_product_angle(v1, v2):
    ang = angle_between(v1, v2)
    if ang == np.inf or ang == np.nan or math.isnan(ang):
        return 0
    return 180 - ang


def pdf_cauchy(u, h, t=1, s=1):
    assert 0 < h < 1
    return 1/(np.pi * np.sqrt(1-(2**(2*h-1)-1)**2)) * 1 / ( (u - (2**(2*h-1) - 1)*(t/s)**h)**2 / ((1 - (2**(2*h-1)-1)**2)*np.sqrt((t/s)**(2*h))) + (np.sqrt((t/s)**(2*h))))


def cdf_cauchy(u, h, t=1, s=1):
    assert 0 < h < 1
    val = np.arctan( (2*(s**h)*u + (t**h)*(2 - 2**(2*h))) / ((t**h) * np.sqrt(2**(2*h+2) - 2**(4*h))) ) / np.pi + 0.5
    return val


def inv_cdf_cauchy(proba, h, t=1, s=1):
    assert 0 < h < 1
    denom = ((t**h) * np.sqrt(2**(2*h+2) - 2**(4*h)))
    a = (2*(s**h)) / denom
    return (np.tan(np.pi * proba - np.pi * 0.5) / a) - ((t**h)*(2 - 2**(2*h)) / (2*(s**h)))  # Fixed precedence bug: `/ 2*(s**h)` -> `/ (2*(s**h))` (Modified by Claude, claude-opus-4-6, Anthropic AI)


def pdf_cauchy_2mixture(x, h1, h2, alpha, beta):
    return beta*(alpha*pdf_cauchy(x, h1) + (1-alpha)*pdf_cauchy(x, h2))


def cdf_cauchy_2mixture(x, h1, h2, alpha, beta):
    return beta*(alpha*cdf_cauchy(x, h1) + (1-alpha)*cdf_cauchy(x, h2))


def pdf_cauchy_3mixture(x, h1, h2, h3, alpha1, alpha2, alpha3, beta):
    return beta*(alpha1*pdf_cauchy(x, h1) + (alpha2)*pdf_cauchy(x, h2) + (alpha3)*pdf_cauchy(x, h3))


def cdf_cauchy_1mixture(x, h1, alpha):
    return alpha*cdf_cauchy(x, h1)


def pdf_cauchy_1mixture(x, h1, alpha):
    return alpha*pdf_cauchy(x, h1)


def cauchy_location(h1, t=1, s=1):
    return (2**(2*h1-1) - 1)*(t/s)**h1


def func_to_fit(func, x, params):
    return func(x, *params)


def func_to_minimise(params, func, x, y):
    y_pred = func_to_fit(func, x, params)
    return np.sum(abs(y_pred - y))


def trajectory_visualization(original_data:pd.DataFrame, analysis_data1:pd.DataFrame, 
                             cutoff:list, pixelmicron:float, resolution_multiplier=80, 
                             roi='', scalebar=True, arrow=False, color_for_roi=False, thickness = 3) -> np.ndarray:
    print("** visualizing trajectories... **")
    scale = resolution_multiplier
    color_maps = {}
    color_maps_plot = {}
    roi_center = None
    angle_circle_radius = 10

    min_x = original_data['x'].min()
    min_y = original_data['y'].min()
    max_x = original_data['x'].max()
    max_y = original_data['y'].max()
    x_width = int(((max_x - min_x) * scale))
    y_width = int(((max_y - min_y) * scale))
    traj_image = np.ones((y_width, x_width, 4)).astype(np.uint8)
    angle_image = np.ones((y_width, x_width, 4)).astype(np.uint8)
    angle_cmap = plt.get_cmap('jet')

    if len(roi) > 4:
        from roifile import ImagejRoi
        contours = ImagejRoi.fromfile(roi).coordinates().astype(np.int32)
        roi_center = (np.array([contours[0][0] - min_x, contours[0][1] - min_y])*scale).astype(int)
        for i in range(len(contours) - 1):
            prev_pt = (np.array([contours[i][0] - min_x, contours[i][1] - min_y])*scale).astype(int)
            next_pt = (np.array([contours[i+1][0] - min_x, contours[i+1][1] - min_y])*scale).astype(int)
            cv2.circle(traj_image, next_pt, thickness+1, (128, 128, 128, 255), -1)
            cv2.circle(angle_image, next_pt, thickness+1, (128, 128, 128, 255), -1)
            roi_center += next_pt
        roi_center = roi_center.astype(float)
        roi_center /= len(contours)

    for traj_idx in tqdm(original_data['traj_idx'].unique(), ncols=120, desc=f'Visualization', unit='trajectory'):
        single_traj = original_data[original_data['traj_idx'] == traj_idx]
        corresponding_ids = analysis_data1['traj_id'] == single_traj['traj_idx'].iloc[0]
        if np.sum(corresponding_ids) > 0:
            traj_color = analysis_data1[analysis_data1['traj_id'] == single_traj['traj_idx'].iloc[0]]['color'].iloc[0]
            state = analysis_data1[analysis_data1['traj_id'] == single_traj['traj_idx'].iloc[0]]['state'].iloc[0]
            if traj_color not in color_maps and traj_color != 'yellow':
                color_maps[traj_color] = str(state)
                color_maps_plot[state] = traj_color
            if cutoff[0] <= len(single_traj) <= cutoff[1]:
                traj_color = mcolors.to_rgb(traj_color)
                traj_color = (int(traj_color[2]*255), int(traj_color[1]*255), int(traj_color[0]*255), 255)  # BGR color for cv2
                pts = np.array([[int((x - min_x) * scale), int((y - min_y) * scale)] for x, y in zip(single_traj['x'], single_traj['y'])], np.int32)
                for i in range(len(pts)-1):
                    prev_pt = pts[i]
                    next_pt = pts[i+1]
                    if arrow:
                        if color_for_roi and roi_center is not None:
                            if np.sqrt(np.sum((next_pt - roi_center)**2)) < np.sqrt(np.sum((prev_pt - roi_center)**2)):
                                cv2.arrowedLine(traj_image, prev_pt, next_pt, (0, 0, 255, 255), thickness)
                            else:
                                cv2.arrowedLine(traj_image, prev_pt, next_pt, (0, 255, 255, 255), thickness)
                            """
                            from module.preprocessing import dot_product_angle
                            angle = dot_product_angle(next_pt - prev_pt, roi_center - prev_pt)
                            if 0 < angle < 90 or 270 < angle < 360:
                                cv2.arrowedLine(image, prev_pt, next_pt, (0, 0, 255), thickness)
                            else:
                                cv2.arrowedLine(image, prev_pt, next_pt, (0, 255, 255), thickness)
                            """
                        else:
                            cv2.arrowedLine(traj_image, prev_pt, next_pt, traj_color, thickness)
                    else:
                        cv2.line(traj_image, prev_pt, next_pt, traj_color, thickness)

                for i in range(len(pts)-2):
                    prev_vec = pts[i+1] - pts[i]
                    next_vec = pts[i+2] - pts[i+1]
                    angle_degree = dot_product_angle(prev_vec, next_vec)
                    angle_color = angle_cmap(angle_degree / 180)
                    cv2.circle(angle_image, center=(pts[i+1][0], pts[i+1][1]), radius=angle_circle_radius,
                               color=(int(angle_color[2]*255), int(angle_color[1]*255), int(angle_color[0]*255), 200), thickness=-1)
                    
    if scalebar:
        cv2.line(traj_image, [int(max(0, x_width - 2*scale - int(scale/pixelmicron))), int(max(0, y_width - 2*scale))], [int(max(0, x_width - 2*scale)) , int(max(0, y_width - 2*scale))], (255, 255, 255, 255), 6)
        cv2.line(angle_image, [int(max(0, x_width - 2*scale - int(scale/pixelmicron))), int(max(0, y_width - 2*scale))], [int(max(0, x_width - 2*scale)) , int(max(0, y_width - 2*scale))], (255, 255, 255, 255), 6)
            
    traj_image = cv2.cvtColor(traj_image, cv2.COLOR_BGRA2RGBA)
    color_maps['yellow'] = 'transitioning'
    color_maps_plot['transitioning'] = 'yellow'
    patches = [mpatches.Patch(color=c,label=color_maps[c]) for c in color_maps]
    return traj_image, angle_image, patches, color_maps, color_maps_plot


import numpy as np
import tifffile
import concurrent.futures
from tqdm import tqdm
from functools import lru_cache
from FreeTrace.module import image_pad
from FreeTrace.module import regression
from FreeTrace.module.data_save import write_localization
from FreeTrace.module.auxiliary import initialization, calibration_3d
from FreeTrace.module.image_module import draw_cross, make_loc_depth_image, read_tif


def region_max_filter2(maps, window_size, thresholds, detect_range=0):
    indices = []
    if detect_range == 0:
        r_start_index = window_size[1] // 2
        col_start_index = window_size[0] // 2
    else:
        r_start_index = detect_range
        col_start_index = detect_range

    for _ in range(2):
        args_map = maps > thresholds.reshape(-1, 1, 1)
        maps = maps * args_map
        img_n, row, col = np.where(args_map == True)
        for n, r, c in zip(img_n, row, col):
            if maps[n, r, c] == np.max(maps[n, max(0, r-r_start_index):min(maps.shape[1]+1, r+r_start_index+1),
                                       max(0, c-col_start_index):min(maps.shape[2]+1, c+col_start_index+1)]) and maps[n, r, c] != 0:
                indices.append([n, r, c])
                maps[n, max(0, r - r_start_index):min(maps.shape[1] + 1, r + r_start_index + 1),
                max(0, c - col_start_index):min(maps.shape[2] + 1, c + col_start_index + 1)] = 0
    return indices


def region_max_filter(maps, window_sizes, thresholds, detect_range=0):
    indices = []
    nb_imgs = maps.shape[1]
    infos = [[] for _ in range(nb_imgs)]
    for i, (hmap, window_size) in enumerate(zip(maps, window_sizes)):
        args_map = hmap > thresholds[:, i].reshape(-1, 1, 1)
        maps[i] = hmap * args_map
        if detect_range == 0:
            r_start_index = window_size[1] // 2
            col_start_index = window_size[0] // 2
        else:
            r_start_index = detect_range
            col_start_index = detect_range
        img_n, row, col = np.where(args_map == True)
        for n, r, c in zip(img_n, row, col):
            if maps[i, n, r, c] == np.max(
                    maps[i, n, max(0, r - r_start_index):min(maps[i].shape[1] + 1, r + r_start_index + 1),
                    max(0, c - col_start_index):min(maps[i].shape[2] + 1, c + col_start_index + 1)]):
                infos[n].append([i, r, c, hmap[n][r][c]])
    maps = np.moveaxis(maps, 0, 1)
    for img_n, info in enumerate(infos):
        mask = np.zeros((maps.shape[2], maps.shape[3])).astype(np.uint8)
        if len(info) > 0:
            info = np.array(info)
            info = info[np.argsort(info[:, 3])[::-1]]
            for mol_info in info:
                if detect_range == 0:
                    extend = (window_sizes[int(mol_info[0])][0] - 1) // 2
                else:
                    extend = detect_range
                row_min = int(max(0, mol_info[1] - extend))
                row_max = int(min(mask.shape[0] - 1, mol_info[1] + extend))
                col_min = int(max(0, mol_info[2] - extend))
                col_max = int(min(mask.shape[1] - 1, mol_info[2] + extend))
                if mask[int(mol_info[1])][int(mol_info[2])] != 1:
                    indices.append([img_n, int(mol_info[1]), int(mol_info[2]), int(window_sizes[int(mol_info[0])][0])])
                    mask[row_min:row_max, col_min:col_max] = 1
    return np.array(indices)


def subtract_pdf(ext_imgs, pdfs, indices, window_size, bg_means, extend):
    for pdf, (n, r, c) in zip(pdfs, indices):
        bg = (np.ones(pdf.shape) * bg_means[n]).reshape(window_size)
        pdf = np.ascontiguousarray(pdf).reshape(window_size)
        row_indice = np.array([r - int((window_size[1]-1)/2), r + int((window_size[1]-1)/2)], dtype=np.intc) + int(extend/2)
        col_indice = np.array([c - int((window_size[0]-1)/2), c + int((window_size[0]-1)/2)], dtype=np.intc) + int(extend/2)
        ext_imgs[n, row_indice[0]:row_indice[1] + 1, col_indice[0]:col_indice[1] + 1] -= pdf
        ext_imgs[n, row_indice[0]:row_indice[1] + 1, col_indice[0]:col_indice[1] + 1] = (
            np.maximum(ext_imgs[n, row_indice[0]:row_indice[1]+1, col_indice[0]:col_indice[1]+1], bg))
        ext_imgs[n] = image_pad.boundary_smoothing(ext_imgs[n], row_indice, col_indice)
    return np.array(ext_imgs)


def gauss_psf(window_sizes, radi):
    gauss_grid_window = []
    for window_size, radius in zip(window_sizes, radi):
        x_subpixel = np.arange(window_size[0]) + .5
        y_subpixel = np.arange(window_size[1]) + .5
        center_x = window_size[0] / 2.
        center_y = window_size[1] / 2.
        base_vals = np.ones((window_size[1], window_size[0], 2)) * np.array([center_x, center_y])
        gauss_psf_vals = np.stack(np.meshgrid(x_subpixel, y_subpixel), -1)
        gauss_psf_vals = np.exp(-(np.sum((gauss_psf_vals - base_vals)**2, axis=2))
                                /(2*(radius**2))) / (np.sqrt(np.pi) * radius)
        gauss_grid_window.append(np.array(gauss_psf_vals, dtype=np.float32))
    return gauss_grid_window


def indice_filtering(indices, window_sizes, img_shape, extend):
    masks = np.zeros(img_shape)
    win_masks = [[[[] for _ in range(img_shape[2])] for _ in range(img_shape[1])] for _ in range(img_shape[0])]
    for indexx, wins in zip(indices[::-1], window_sizes[::-1]):
        for index in indexx:
            masks[index[0], index[1], index[2]] += 1
            win_masks[index[0]][index[1]][index[2]].append(wins[0])
    ret_indices = check_masks_overlaps(masks, win_masks, extend, window_sizes)
    return ret_indices


def check_masks_overlaps(masks, window_masks, extend, window_sizes):
    nb_window_sizes = len(window_sizes)
    w_size_dict = {ws[0]: i for i, ws in enumerate(window_sizes)}
    all_groups = []
    for img_n, (mask, window_mask) in enumerate(zip(masks, window_masks)):
        groups = []
        overlay_mask = np.zeros_like(mask, dtype=np.uint8)
        rs, cs = np.where(mask >= 1)
        if len(rs) == 0:
            continue

        coords = np.vstack((rs, cs)).T
        while 1:
            group = [[] for _ in range(nb_window_sizes)]
            selected_args = []
            piv_coord = coords[0]

            for window_size in window_mask[piv_coord[0]][piv_coord[1]]:
                group[w_size_dict[window_size]].append([img_n, piv_coord[0] + extend, piv_coord[1] + extend, window_size])
                row_min = max(0, piv_coord[0] - (window_size // 2))
                row_max = min(mask.shape[0] - 1, piv_coord[0] + (window_size // 2))
                col_min = max(0, piv_coord[1] - (window_size // 2))
                col_max = min(mask.shape[1] - 1, piv_coord[1] + (window_size // 2))
                overlay_mask[row_min: row_max+1, col_min: col_max+1] += 1

            explorer_coords = coords[1:].copy()
            while 1:
                group_added = 0
                for selected_index, coord in enumerate(explorer_coords):
                    if selected_index in selected_args:
                        continue
                    if overlay_mask[coord[0]][coord[1]] >= 1:
                        selected_args.append(selected_index)
                        for window_size in window_mask[coord[0]][coord[1]]:
                            row_min = max(0, coord[0] - (window_size // 2))
                            row_max = min(mask.shape[0] - 1, coord[0] + (window_size // 2))
                            col_min = max(0, coord[1] - (window_size // 2))
                            col_max = min(mask.shape[1] - 1, coord[1] + (window_size // 2))
                            overlay_mask[row_min: row_max + 1, col_min: col_max + 1] += 1
                            group[w_size_dict[window_size]].append([img_n, coord[0] + extend, coord[1] + extend, window_size])
                            group_added += 1

                if group_added == 0:
                    groups.append(group)
                    break
            selected_args = np.array(selected_args) + 1
            selected_args = np.append(selected_args, 0).astype(np.uint32)
            coords = np.delete(coords, selected_args, axis=0)
            if len(coords) == 0:
                break
        all_groups.extend(groups)
    return all_groups


def localization(imgs: np.ndarray, bgs, f_gauss_grids, b_gauss_grids, *args):
    pass_to_multi = False
    single_winsizes = args[0]
    single_thresholds = args[2]
    multi_winsizes = args[3]
    multi_thresholds = args[5]
    p0=args[6]
    shift=args[7]
    decomp_n=args[8]
    deflation_loop_backward = args[10]
    extend = multi_winsizes[-1][0]*4

    coords = [[] for _ in range(imgs.shape[0])]
    reg_pdfs = [[] for _ in range(imgs.shape[0])]
    reg_infos = [[] for _ in range(imgs.shape[0])]
    bg_means = bgs[multi_winsizes[0][0]][:, 0]
    extended_imgs = np.zeros((imgs.shape[0], imgs.shape[1] + extend, imgs.shape[2] + extend), dtype=np.float32)
    extended_imgs[:, int(extend/2):int(extend/2) + imgs.shape[1], int(extend/2):int(extend/2) + imgs.shape[2]] += imgs
    extended_imgs_copy = extended_imgs.copy()
    extended_imgs = image_pad.add_block_noise(extended_imgs, extend)

    while True:
        all_crop_imgs = {ws[0]: None for ws in single_winsizes}
        win_s_dict = {}
        for ws in single_winsizes:
            win_s_dict[ws[0]] = []

        if pass_to_multi:
            for df_loop in range(deflation_loop_backward):
                h_maps = []
                for g_grid, window_size in zip(b_gauss_grids, multi_winsizes):
                    if CUDA:
                        crop_imgs, cp_type = gpu_module.image_cropping(extended_imgs, extend, window_size[0], window_size[1], shift=shift)
                        bg_squared_sums = window_size[0] * window_size[1] * bg_means**2
                        c = gpu_module.likelihood(crop_imgs, g_grid, bg_squared_sums, bg_means, window_size[0], window_size[1])
                        if cp_type:
                            crop_imgs = crop_imgs.get()
                    else:
                        crop_imgs = image_pad.image_cropping(extended_imgs, extend, window_size[0], window_size[1], shift=shift)
                        bg_squared_sums = window_size[0] * window_size[1] * bg_means**2
                        c = image_pad.likelihood(crop_imgs, g_grid, bg_squared_sums, bg_means, window_size[0], window_size[1])

                    h_map = np.array(image_pad.mapping(c, imgs.shape[0], imgs.shape[1], imgs.shape[2], shift))
                    h_map = h_map * (multi_winsizes[0][0]**2 / window_size[0]**2)
                    h_maps.append(h_map)
                h_maps = np.array(h_maps)

                """
                plt.figure()
                plt.imshow(imgs[0])
                plt.figure()
                plt.imshow(extended_imgs[0])
                for hm in h_maps:
                    plt.figure()
                    plt.imshow(hm[0], vmin=0., vmax=1.)
                plt.figure()
                plt.imshow(extended_imgs[0])
                plt.show()
                """

                back_indices = [[] for _ in range(multi_thresholds.shape[1])]
                for backward_index in range(multi_thresholds.shape[1]-1, -1, -1):
                    back_indices[backward_index] = region_max_filter2(h_maps[backward_index], multi_winsizes[backward_index],
                                                                      multi_thresholds[:, backward_index], detect_range=shift)
                reregress_indice = indice_filtering(back_indices, multi_winsizes, imgs.shape, int(extend/2))
                regress_imgs_copy = extended_imgs.copy()
                for regress_comp_set in reregress_indice:
                    loss_vals = []
                    selected_dt = []
                    for win_s_set in regress_comp_set:
                        regress_imgs = []
                        bg_regress = []
                        for regress_index in win_s_set:
                            ws = regress_index[3]
                            regress_imgs.append(regress_imgs_copy[regress_index[0],
                                                regress_index[1] - int((regress_index[3] - 1) / 2):regress_index[1] + int(
                                                    (regress_index[3] - 1) / 2) + 1,
                                                regress_index[2] - int((regress_index[3] - 1) / 2):regress_index[2] + int(
                                                    (regress_index[3] - 1) / 2) + 1])
                            bg_regress.append(bgs[ws][regress_index[0]])
                        regress_imgs = np.array(regress_imgs)

                        if len(regress_imgs) > 0:
                            ############ FORCE BOUNDARY TO MIN ############
                            #min_tmp = np.min(regress_imgs, axis=(1, 2)).reshape(-1, 1)
                            #regress_imgs[:,:,0] = min_tmp
                            #regress_imgs[:,:,-1] = min_tmp
                            #regress_imgs[:,0,:] = min_tmp
                            #regress_imgs[:,-1,:] = min_tmp
                            ###############################################
                            pdfs, xs, ys, x_vars, y_vars, amps, rhos = image_regression(regress_imgs, bg_regress,
                                                                                        (ws, ws), p0=p0, decomp_n=decomp_n)

                            penalty = 0
                            for x_var, y_var, rho, dx, dy in zip(x_vars, y_vars, rhos, xs, ys):
                                if x_var < 0 or y_var < 0 or x_var > 3*ws or y_var > 3*ws or rho > 1 or rho < -1:
                                    penalty += 1e6
                            regressed_imgs = []
                            for regress_index, dx, dy in zip(win_s_set, xs, ys):
                                reged_img = regress_imgs_copy[regress_index[0],
                                            regress_index[1] - int((ws - 1) / 2) + int(np.round(dy)):
                                            regress_index[1] + int((ws - 1) / 2) + int(np.round(dy)) + 1,
                                            regress_index[2] - int((ws - 1) / 2) + int(np.round(dx)):
                                            regress_index[2] + int((ws - 1) / 2) + int(np.round(dx)) + 1]
                                if reged_img.shape == (ws, ws):
                                    regressed_imgs.append(reged_img)
                            regressed_imgs = np.array(regressed_imgs)
                            selected_dt.append([pdfs, xs, ys, x_vars, y_vars, rhos, amps])
                            if regressed_imgs.shape != (pdfs.reshape(regress_imgs.shape)).shape:
                                ## x_var or y_var is (-)
                                loss_vals.append(penalty)
                            else:
                                loss = np.mean(np.sort(np.mean(np.log(abs(regressed_imgs - pdfs.reshape(regress_imgs.shape))), axis=0).
                                                       flatten())[::-1][:multi_winsizes[0][0] * multi_winsizes[0][1]]) + penalty
                                loss_vals.append(loss)
                        else:
                            selected_dt.append([0, 0, 0, 0, 0, 0, 0])
                            loss_vals.append(1e3)

                    if np.sum(np.array(loss_vals) < 0.) >= 1:
                        selec_arg = np.argmin(loss_vals)
                        pdfs, xs, ys, x_vars, y_vars, rhos, amps = selected_dt[selec_arg]
                        infos = np.array(regress_comp_set[selec_arg])
                        ns = infos[:, 0]
                        rs = infos[:, 1]
                        cs = infos[:, 2]
                        ws = infos[:, 3][0]
                        for n, r, c, dx, dy, pdf, x_var, y_var, rho, amp in zip(ns, rs, cs, xs, ys, pdfs, x_vars, y_vars, rhos, amps):
                            r -= int(extend/2)
                            c -= int(extend/2)
                            if r+dy <= -1 or r+dy >= imgs.shape[1] or c+dx <= -1 or c+dx >= imgs.shape[2]:
                                continue
                            x_coord = max(0, min(r+dy, imgs.shape[1]-1))
                            y_coord = max(0, min(c+dx, imgs.shape[2]-1))
                            z_coord = 0.0
                            coords[n].append([x_coord, y_coord, z_coord])
                            reg_pdfs[n].append(pdf)
                            reg_infos[n].append([x_var, y_var, rho, amp])
                        if df_loop < deflation_loop_backward - 1:
                            del_indices = np.round(np.array([ns, rs+ys, cs+xs])).astype(int).T
                            extended_imgs = subtract_pdf(extended_imgs, pdfs, del_indices, (ws, ws), bg_means, extend=0)
            return coords, reg_pdfs, reg_infos

        else:
            h_maps = []
            #total_before_time = timer()
            for g_grid, window_size in zip(f_gauss_grids, single_winsizes):      
                if CUDA:
                    crop_imgs, cp_type = gpu_module.image_cropping(extended_imgs, extend, window_size[0], window_size[1], shift=shift)
                    bg_squared_sums = window_size[0] * window_size[1] * bg_means**2
                    c = gpu_module.likelihood(crop_imgs, g_grid, bg_squared_sums, bg_means, window_size[0], window_size[1])
                    if cp_type:
                        crop_imgs = crop_imgs.get()
                else:
                    crop_imgs = np.array(image_pad.image_cropping(extended_imgs, extend, window_size[0], window_size[1], shift=shift), dtype=np.float32)
                    bg_squared_sums = window_size[0] * window_size[1] * bg_means**2
                    c = np.array(image_pad.likelihood(crop_imgs, g_grid, bg_squared_sums, bg_means, int(window_size[0]), int(window_size[1])), dtype=np.float32)

                all_crop_imgs[window_size[0]] = crop_imgs
                h_map = np.array(image_pad.mapping(c, imgs.shape[0], imgs.shape[1], imgs.shape[2], shift))
                h_map = h_map * (single_winsizes[0][0]**2 / window_size[0]**2)
                h_maps.append(h_map)
            h_maps = np.array(h_maps)
            #print(f'{"h_map calcul":<35}:{(timer() - total_before_time):.2f}s')
            
            """
            import matplotlib.pyplot as plt
            import glob
            print(extended_imgs.shape, h_maps.shape)
            print('threshold:, ',single_thresholds)
            for get in range(extended_imgs.shape[0]):
                fig, axs = plt.subplots(1, 2, figsize=(8, 4))
                axs[0].imshow(extended_imgs[get], cmap='gray')
                axs[1].imshow(h_maps[0][get], cmap='gray')
                plt_idx = len(glob.glob('tempo/*.png'))
                axs[0].set_xticks([], [])
                axs[0].set_yticks([], [])
                axs[1].set_xticks([], [])
                axs[1].set_yticks([], [])
                plt.tight_layout()
                plt.savefig(f'tempo/{plt_idx}.png')
                plt.close('all')
            """
            
            indices = region_max_filter(h_maps, single_winsizes, single_thresholds, detect_range=shift)
            if len(indices) != 0:
                for n, r, c, ws in indices:
                    win_s_dict[ws].append([all_crop_imgs[ws][n]
                                           [int((r//shift) * (imgs.shape[2]//shift) + (c//shift))],
                                           bgs[ws][n], n, r, c])
                ws = single_winsizes[0][0]
                if len(win_s_dict[ws]) != 0:
                    err_indice = []
                    nb_calc = len(win_s_dict[ws])
                    win_s_dict[ws] = np.array(win_s_dict[ws], dtype=np.object_)
                    regress_imgs = np.array([win_s_dict[ws][i][0] for i in range(nb_calc)])
                    bg_regress = np.array([win_s_dict[ws][i][1] for i in range(nb_calc)])
                    ns = np.array([win_s_dict[ws][i][2] for i in range(nb_calc)])
                    rs = np.array([win_s_dict[ws][i][3] for i in range(nb_calc)])
                    cs = np.array([win_s_dict[ws][i][4] for i in range(nb_calc)])

                    # Warning
                    #min_local_val = np.array([np.min(regress_imgs, axis=1) for _ in range(ws)]).T
                    regress_imgs = regress_imgs.reshape(-1, ws, ws)
                    #regress_imgs[:, :, 0] = min_local_val
                    #regress_imgs[:, :, -1] = min_local_val
                    #regress_imgs[:, 0, :] = min_local_val
                    #regress_imgs[:, -1, :] = min_local_val
                    

                    pdfs, xs, ys, x_vars, y_vars, amps, rhos = image_regression(regress_imgs, bg_regress,
                                                                                (ws, ws), p0=p0, decomp_n=decomp_n)
                    for err_i, (x_var, y_var, rho) in enumerate(zip(x_vars, y_vars, rhos)):
                        if x_var < 0 or y_var < 0 or x_var > 3*ws or y_var > 3*ws or rho > 1 or rho < -1:
                            err_indice.append(err_i)

                    if len(err_indice) == len(pdfs):
                        #print(f'IMPOSSIBLE REGRESSION(MINUS VAR): {err_indice}\nWindow_size:{ws}')
                        pass_to_multi = True
                    else:
                        pdfs = np.delete(pdfs, err_indice, 0)
                        xs = np.delete(xs, err_indice, 0)
                        ys = np.delete(ys, err_indice, 0)
                        x_vars = np.delete(x_vars, err_indice, 0)
                        y_vars = np.delete(y_vars, err_indice, 0)
                        ns = np.delete(ns, err_indice, 0)
                        rs = np.delete(rs, err_indice, 0)
                        cs = np.delete(cs, err_indice, 0)
                        rhos = np.delete(rhos, err_indice, 0)  # MODIFIED 11/03/2026
                        amps = np.delete(amps, err_indice, 0)  # MODIFIED 11/03/2026
                        for n, r, c, dx, dy, pdf, x_var, y_var, rho, amp in zip(ns, rs, cs, xs, ys, pdfs, x_vars, y_vars, rhos, amps):
                            if r+dy <= -1 or r+dy >= imgs.shape[1] or c+dx <= -1 or c+dx >= imgs.shape[2]:
                                continue
                            x_coord = max(0, min(r+dy, imgs.shape[1]-1))
                            y_coord = max(0, min(c+dx, imgs.shape[2]-1))
                            z_coord = 0.0
                            coords[n].append([x_coord, y_coord, z_coord])
                            reg_pdfs[n].append(pdf)
                            reg_infos[n].append([x_var, y_var, rho, amp])
                        del_indices = np.round(np.array([ns, rs+ys, cs+xs])).astype(int).T

                        #extended_imgs_copy = extended_imgs.copy()
                        extended_imgs = subtract_pdf(extended_imgs, pdfs, del_indices, (ws, ws), bg_means, extend)
                        """
                        print('\ntreshold:', single_thresholds)
                        plt.figure('before subtraction', figsize=(8, 8))
                        #print(del_indices)
                        for n, x, y in del_indices:
                            if n==get:
                                #print(n, x, y, extend)
                        
                                plt.scatter(y +int(extend/2), x+int(extend/2) ,c='red')
       
                        plt.imshow(extended_imgs_copy[get])
                        extended_imgs = subtract_pdf(extended_imgs, pdfs, del_indices, (ws, ws), bg_means, extend)
                        plt.figure('after subtraction', figsize=(8, 8))
                        plt.imshow(extended_imgs[get])
                        plt.show()
                        """
                        
                        
            pass_to_multi = True  # forbid deflation loop.
            #if np.allclose(extended_imgs_copy, extended_imgs, atol=1e-2) or len(indices) == 0 or single_winsizes[0][0] not in indices[:, 3]:
            #    pass_to_multi = True


@lru_cache(maxsize=8)
def initial_pdf(shape_0, shape_1, shape_2):
    return np.ones((shape_0, shape_1, shape_2), dtype=np.float32)


def bi_variate_normal_pdf(xy, cov, mu, normalization=True):
    init_pdf = initial_pdf(cov.shape[0], xy.shape[0] ,xy.shape[1])
    init_pdf *= (xy - mu)
    if normalization:
        return (np.exp((-1./2) * np.sum(init_pdf @ np.linalg.inv(cov) * init_pdf, axis=2))
                / (2 * np.pi * np.sqrt(np.linalg.det(cov).reshape(-1, 1))))
    else:
        return (np.exp((-1./2) * np.sum(init_pdf @ np.linalg.inv(cov) * init_pdf, axis=2)))


@lru_cache(maxsize=8)
def quantification(window_size):
    x = np.arange(-(window_size[0]-1)/2, (window_size[0]+1)/2)
    y = np.arange(-(window_size[1]-1)/2, (window_size[1]+1)/2)
    xv, yv = np.meshgrid(x, y, sparse=True)
    grid = np.stack(np.meshgrid(xv, yv), -1).reshape(window_size[0] * window_size[1], 2)
    return grid.astype(np.float32)


@lru_cache(maxsize=8)
def make_x_grid(win_sizes):
    return (np.array([list(np.arange(-int(win_sizes[0]/2), int((win_sizes[0]/2) + 1), 1))] * win_sizes[1], dtype=np.float32).reshape(-1, win_sizes[0] * win_sizes[1]))


@lru_cache(maxsize=8)
def make_y_grid(win_sizes):
    return (np.array([[y] * win_sizes[0] for y in range(-int(win_sizes[1]/2), int((win_sizes[1]/2) + 1), 1)], dtype=np.float32).reshape(-1, win_sizes[0] * win_sizes[1]))


def image_regression(imgs, bgs, window_size, p0, decomp_n, amp=0, repeat=5):
    imgs = np.array(imgs).reshape([-1, window_size[0] * window_size[1]])
    bgs = np.array(bgs).reshape([-1, window_size[0] * window_size[1]])
    p0 = np.array(p0, dtype=np.float32)
    x_grid = make_x_grid(window_size)
    y_grid = make_y_grid(window_size)
    grid = quantification(window_size)
    coefs = regression.guo_algorithm(imgs, bgs, p0=p0, xgrid=x_grid, ygrid=y_grid, 
                                     window_size=window_size, repeat=repeat, decomp_n=decomp_n)
    variables, err_indices = regression.unpack_coefs(coefs, window_size)

    if len(err_indices) > 0:
        coefs = regression.guo_algorithm(imgs, bgs, p0=p0, xgrid=x_grid, ygrid=y_grid,
                                         window_size=window_size, repeat=repeat+1, decomp_n=decomp_n)
        variables, err_indices = regression.unpack_coefs(coefs, window_size)
    variables = np.array(variables).T
    cov_mat = np.array([variables[:, 0], variables[:, 4] * np.sqrt(variables[:, 0]) * np.sqrt(variables[:, 2]),
                        variables[:, 4] * np.sqrt(variables[:, 0] * np.sqrt(variables[:, 2])), variables[:, 2]]
                       ).T.reshape(variables.shape[0], 2, 2)
    pdfs = bi_variate_normal_pdf(grid, cov_mat, mu=np.array([0, 0]), normalization=False)
    pdfs = variables[:, 5].reshape(-1, 1) * pdfs + bgs
    for err_i in err_indices:
        variables[err_i][0] = -100
        variables[err_i][2] = -100
    return pdfs, variables[:, 1], variables[:, 3], variables[:, 0], variables[:, 2], variables[:, 5], variables[:, 4]


def make_red_circles(imgs, localized_xys, hstack=False):
    if hstack:
        original_imgs = imgs.copy()
    stacked_imgs = []
    for img_n, coords in enumerate(localized_xys):
        xy_cum = []
        for center_coord in coords:
            x, y = int(round(center_coord[0])), int(round(center_coord[1]))
            if (x, y) in xy_cum:
                imgs[img_n] = draw_cross(imgs[img_n], x, y, (0, 0, 1))
            else:
                imgs[img_n] = draw_cross(imgs[img_n], x, y, (1, 0, 0))
            xy_cum.append((x, y))
        if hstack:
            stacked_imgs.append(np.hstack((original_imgs[img_n], imgs[img_n])))
        else:
            stacked_imgs.append(imgs[img_n])
    return np.array(stacked_imgs, dtype=np.uint8)


def visualilzation(output_dir, images, localized_xys, hstack=False):
    orignal_imgs_3ch = np.array([images * 255, images * 255, images * 255], dtype=np.uint8)
    orignal_imgs_3ch = np.ascontiguousarray(np.moveaxis(orignal_imgs_3ch, 0, 3))
    stacked_img = make_red_circles(orignal_imgs_3ch, localized_xys, hstack)
    tifffile.imwrite(f'{output_dir}_locvideo.tiff', data=stacked_img, imagej=True)


def background(imgs, window_sizes, alpha):
    imgs = imgs / np.max(imgs, axis= (1,2)).reshape(-1, 1, 1)
    bins = 0.01
    bgs = {}
    bg_means = []
    bg_stds = []
    bg_intensities = (imgs.reshape(imgs.shape[0], imgs.shape[1] * imgs.shape[2]) * 100).astype(np.uint8) / 100
    for i in range(len(bg_intensities)):
        args = np.arange(len(bg_intensities[i]))
        post_mask_args = args.copy()
        for _ in range(3):
            it_hist, bin_width = np.histogram(bg_intensities[i][post_mask_args], bins=np.arange(0, np.max(bg_intensities[i][post_mask_args]) + bins, bins))
            if len(it_hist) < 1:
                print('Errors on images, please check images again whether it contains an empty black-image. If not, contact the author.')
                break
            mask_sums_mode = (np.argmax(it_hist) * bins + (bins / 2))
            mask_std = np.std(bg_intensities[i][post_mask_args])
            post_mask_args = np.array([arg for arg, val in zip(args, bg_intensities[i]) if
                                       (mask_sums_mode - 3. * mask_std) < val < (mask_sums_mode + 3. * mask_std)])
        it_data = bg_intensities[i][post_mask_args]
        bg_means.append(np.mean(it_data))
        bg_stds.append(np.std(it_data))
    bg_means = np.array(bg_means)
    bg_stds = np.array(bg_stds)

    for window_size in window_sizes:
        bg = np.ones((bg_intensities.shape[0], window_size[0] * window_size[1]), dtype=np.float32)
        bg *= bg_means.reshape(-1, 1)
        bgs[window_size[0]] = bg

    thresholds = np.array(1/(bg_means**2 / bg_stds**2)) * 2.0
    thresholds = np.maximum(thresholds, np.ones_like(thresholds) * 1.0)
    for th_i in range(len(thresholds)):
        if np.isnan(thresholds[th_i]):
            thresholds[th_i] = 1.0
    return bgs, thresholds * alpha


def intensity_distribution(images, reg_pdfs, xyz_coords, reg_infos, sigma=3.5):
    new_pdfs = []
    new_coords = []
    new_infos = []
    for img_n, (pdfs, xyz_coord, infos) in enumerate(zip(reg_pdfs, xyz_coords, reg_infos)):
        if len(pdfs) < 1:
            continue
        new_pdf_tmp = pdfs.copy()
        new_xyz_coord_tmp = xyz_coord.copy()
        new_reg_tmp = infos.copy()
        max_pdf_vals = []

        for pdf, xy, info in zip(pdfs, xyz_coord, infos):
            max_pdf_vals.append(np.max(images[img_n,
                                       max(0, int(np.round(xy[0])) - 1): min(images[img_n].shape[0], int(np.round(xy[0])) + 2),
                                       max(0, int(np.round(xy[1])) - 1): min(images[img_n].shape[1], int(np.round(xy[1])) + 2)]))

        max_pdf_vals = np.array(max_pdf_vals)
        bin_edgs = np.arange(0, np.max(max_pdf_vals) + 0.05, 0.025)
        max_pdf_vals_hist = np.histogram(max_pdf_vals, bins=bin_edgs)
        mode_sigma = (bin_edgs[:-1] + 0.025)[np.argmax(max_pdf_vals_hist[0])] + sigma * np.std(max_pdf_vals)

        for i, max_pdf_val in enumerate(max_pdf_vals):
            if max_pdf_val > mode_sigma:
                new_pdf_tmp.append(pdfs[i])
                new_xyz_coord_tmp.append(xyz_coord[i])
                new_reg_tmp.append(infos[i])
        new_pdfs.append(new_pdf_tmp)
        new_coords.append(new_xyz_coord_tmp)
        new_infos.append(new_reg_tmp)
    return new_pdfs, new_coords, new_infos


def params_gen(win_s):
    assert type(win_s) is int
    if win_s < 5:
        win_s = 5
    if win_s % 2 == 0:
        win_s += 1

    single_winsizes = [(win_s, win_s)]
    multi_winsizes = [(win_s-2, win_s-2), (win_s, win_s), (win_s+2, win_s+2)]
    single_radius = [((r[0]//2) / 2.) for r in single_winsizes]
    multi_radius = [((r[0]//2) / 2.) for r in multi_winsizes]
    return single_winsizes, single_radius, multi_winsizes, multi_radius


def main_process(imgs, forward_gauss_grids, backward_gauss_grids, *args):
    args = list(args)
    if CUDA:
        bgs, thresholds = gpu_module.background(imgs, window_sizes=args[3], alpha=args[9])
    else:
        bgs, thresholds = background(imgs, window_sizes=args[3], alpha=args[9])

    if args[2] is None:
        args[2] = np.array([thresholds for _ in range(len(args[0]))]).T
    else:
        args[2] = np.ones((len(imgs), len(args[0]))) * args[2]
    if args[5] is None:
        args[5] = np.array([thresholds for _ in range(len(args[3]))]).T
    else:
        args[5] = np.ones((len(imgs), len(args[3]))) * args[5]

    xyz_coord, pdf, info = localization(imgs, bgs, forward_gauss_grids, backward_gauss_grids, *args)
    return xyz_coord, pdf, info


def run(input_video_path:str, output_path:str, window_size=7, threshold=1.0, deflation=0, sigma=4.0, shift=1, z_calib=None, # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
        gpu_on=True, save_video=False, verbose=False, batch=False, realtime_visualization=False, return_state=0, batch_size=0): # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
    
    global VERBOSE
    global BATCH
    global MEM_SIZE 
    global WINSIZE 
    global THRES_ALPHA 
    global DEFLATION_LOOP_IN_BACKWARD
    global CUDA
    global P0 
    global GAUSS_SEIDEL_DECOMP
    global CORE 
    global PARALLEL
    global PBAR
    global BINARY_THRESHOLDS
    global MULTI_THRESHOLDS
    global gpu_module


    loc_output_path = f'{output_path}/{input_video_path.split("/")[-1].split(".tif")[0]}'
    WINSIZE = window_size
    THRES_ALPHA = threshold
    DEFLATION_LOOP_IN_BACKWARD = deflation
    VERBOSE = verbose
    BATCH = batch
    P0 = [1.5, 0., 1.5, 0., 0., 0.5]
    GAUSS_SEIDEL_DECOMP = 1
    CORE = 4
    PARALLEL = False
    BINARY_THRESHOLDS = None
    MULTI_THRESHOLDS = None


    images = read_tif(input_video_path)
    CUDA, _ = initialization(gpu_on, ptype=0, verbose=VERBOSE, batch=BATCH)
    if CUDA:
        from FreeTrace.module import gpu_module
        MEM_SIZE = gpu_module.get_gpu_mem_size()
    if batch_size > 0: # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
        DIV_Q = batch_size
    elif CUDA:
        DIV_Q = int(64 * 512 * 512 / images.shape[1] / images.shape[2] * (7**2 / WINSIZE**2) * MEM_SIZE / 24 * shift**2)
    else:
        DIV_Q = min(50, int(2.7 * 4194304 / images.shape[1] / images.shape[2] * (7**2 / WINSIZE**2)))
    DIV_Q = max(DIV_Q, 1) # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
    if WINSIZE % 2 == 0:
        WINSIZE += 1

    dim = 2
    xyz_coords = []
    reg_pdfs = []
    reg_infos = []
    SINGLE_WINSIZES, SINGLE_RADIUS, MULTI_WINSIZES, MULTI_RADIUS = params_gen(WINSIZE)
    forward_gauss_grids = gauss_psf(SINGLE_WINSIZES,SINGLE_RADIUS)
    backward_gauss_grids = gauss_psf(MULTI_WINSIZES, MULTI_RADIUS)


    realtime_obj = None
    if realtime_visualization:
        from FreeTrace.module.image_module import RealTimePlot
        realtime_obj = RealTimePlot(f'Localization : {input_video_path}', show_frame=True)
        realtime_obj.turn_on()


    np.random.seed(42)  # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-14
    if VERBOSE:
        PBAR = tqdm(total=len(images), desc="Localization", unit=f"frame", ncols=120)
    if PARALLEL:
        for div_q in range(0, len(images), CORE * DIV_Q):
            with concurrent.futures.ProcessPoolExecutor() as executor:
                executors = {i: None for i in range(CORE)}
                for cc in range(CORE):
                    if div_q + cc*DIV_Q < len(images):
                        future = executor.submit(main_process, images[div_q + cc*DIV_Q: div_q + cc*DIV_Q + DIV_Q],
                                                 forward_gauss_grids, backward_gauss_grids,
                                                 SINGLE_WINSIZES, SINGLE_RADIUS, BINARY_THRESHOLDS,
                                                 MULTI_WINSIZES, MULTI_RADIUS, MULTI_THRESHOLDS,
                                                 P0, shift, GAUSS_SEIDEL_DECOMP, THRES_ALPHA, DEFLATION_LOOP_IN_BACKWARD)
                        executors[cc] = future
                for wait_ in executors:
                    if type(executors[wait_]) is concurrent.futures.Future:
                        xyz_coord, pdf, info = executors[wait_].result()
                        xyz_coords.extend(xyz_coord)
                        reg_pdfs.extend(pdf)
                        reg_infos.extend(info)
    else:
        for div_q in range(0, len(images), DIV_Q):
            xyz_coord, pdf, info = main_process(images[div_q:div_q+DIV_Q], forward_gauss_grids, backward_gauss_grids,
                                               SINGLE_WINSIZES, SINGLE_RADIUS, BINARY_THRESHOLDS,
                                               MULTI_WINSIZES, MULTI_RADIUS, MULTI_THRESHOLDS,
                                               P0, shift, GAUSS_SEIDEL_DECOMP, THRES_ALPHA, DEFLATION_LOOP_IN_BACKWARD)
            # Coord date is stored as y, x, z
            xyz_coords.extend(xyz_coord)
            reg_pdfs.extend(pdf)
            reg_infos.extend(info)
            if realtime_visualization:
                realtime_obj.put_into_queue((images[div_q:div_q+DIV_Q], xyz_coord, div_q), mod_n=2) 

            if VERBOSE and len(images[div_q:div_q+DIV_Q]) == DIV_Q:
                PBAR.update(DIV_Q)
            elif VERBOSE:
                PBAR.update(len(images) % DIV_Q)

    if VERBOSE:
        PBAR.close()

    #reg_pdfs, xyz_coords, reg_infos = intensity_distribution(images, reg_pdfs, xyz_coords, reg_infos, sigma=sigma)
    if z_calib is not None:
        dim = 3
        xyz_coords = calibration_3d(xyz_coords, reg_infos, z_calib)

    if save_video:
        print(f'Visualizing localizations...')
        visualilzation(loc_output_path, images, xyz_coords)
    if realtime_obj is not None:
        realtime_obj.turn_off()

    if return_state != 0:
        return_state.value = 1

    write_localization(loc_output_path, xyz_coords, reg_pdfs, reg_infos)
    make_loc_depth_image(loc_output_path, xyz_coords, multiplier=4, winsize=WINSIZE, resolution=2, dim=dim)

    return True


def run_process(input_video_path:str, output_path:str, window_size=7, threshold=1.0, deflation=0, sigma=4.0, shift=1, z_calib=None, # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
                gpu_on=True, save_video=False, verbose=False, batch=False, realtime_visualization=False, batch_size=0) -> bool: # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
    """
    Create a process to localize the positions of particles from video.

    @params
        input_video_path:
        Path of input video. Currently we only accept .tiff format. (video.tiff)

        output_path:
        Path to save the output files. (video_loc.csv and supplementary outputs depending on the visualization options will be saved in this path)
        
        window_size (pixel):
        Sliding window length at pixel scale to search the signals of particles in each frame of video. 
        
        threshold:
        Threshold multiplier to determine the existence of particle inside a sliding window. Low value increases the detection rate. Base threshold is calculated with SNR of each frame. Recommended range is [0.5, 1.0].
        
        deflation:
        Deflation loop to search particles iteratively. (deprecated)
        
        sigma:
        Standard deviation to detect the overlapped particles. (deprecated)
        
        shift:
        Value of pixel to shift the sliding window for the detction of particles.

        z_calib:
        Write y_variance - x_varince on z coordinates. (developping)
        
        gpu_on:
        Accelerate the computation of localization using GPU.
        
        save_video:
        Save a video of localized particle result.
        
        verbose:
        Print the progress.
        
        realtime_visualization:
        Real time visualization of progress.

    @return
        return:
        It returns True if the localization of particles is finished succesfully, False otherwise.
        The unit of saved particle coordinates are pixel.
    """
    
    from multiprocessing import Process, Value
    return_state = Value('b', 0)
    options = {
        'window_size': window_size,
        'threshold': threshold,
        'deflation': deflation,
        'sigma': sigma,
        'shift': shift,
        'z_calib': z_calib,
        'gpu_on': gpu_on,
        'save_video': save_video,
        'verbose': verbose,
        'batch': batch,
        'realtime_visualization': realtime_visualization,
        'return_state': return_state,
        'batch_size': batch_size # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-15
    }

    p = Process(target=run, args=(input_video_path, output_path),  kwargs=options)
    try:
        p.start()
        p.join()
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating childs")
        p.terminate()
        p.join()
    finally:
        p.close()
    return return_state.value



# Modified by Claude (claude-opus-4-6, Anthropic AI) — added package init, model download on first import
"""FreeTrace — single-molecule tracking with fBm inference."""
__version__ = "1.6.1.0"  # Modified by Claude (claude-opus-4-6, Anthropic AI) - 2026-03-18

from FreeTrace.model_downloader import ensure_models as _ensure_models
_ensure_models()
del _ensure_models



from setuptools import setup
from Cython.Build import cythonize
import os
import glob
import shutil
import re


freetrace_path = ''
freetrace_path += 'FreeTrace'.join(re.split(r'FreeTrace', __file__)[:-1]) + 'FreeTrace'
freetrace_path = freetrace_path.replace('\\', '/')
build_path = f"{freetrace_path}/module/cython_build"
print(f"Current build path: {build_path}")


setup(
    name='FreeTrace app',
    ext_modules=cythonize([f"{build_path}/image_pad.pyx", f"{build_path}/regression.pyx", f"{build_path}/cost_function.pyx"], language_level = "3", annotate=True),
)


source_file = glob.glob(f"{os.getcwd()}/cost_function*")[0]
extens = source_file.split(".")[-1]
destination_path = f"{freetrace_path}/module/cost_function.{extens}"
shutil.copy(source_file, destination_path)
os.remove(source_file)


source_file = glob.glob(f"{os.getcwd()}/image_pad*")[0]
extens = source_file.split(".")[-1]
destination_path = f"{freetrace_path}/module/image_pad.{extens}"
shutil.copy(source_file, destination_path)
os.remove(source_file)


source_file = glob.glob(f"{os.getcwd()}/regression*")[0]
extens = source_file.split(".")[-1]
destination_path = f"{freetrace_path}/module/regression.{extens}"
shutil.copy(source_file, destination_path)
os.remove(source_file)


directory_to_remove = f"{build_path}/build"
if os.path.exists(directory_to_remove):
    try:
        shutil.rmtree(directory_to_remove)
    except OSError as e:
        print(f"Error: {e.filename} - {e.strerror}.")


# Install Cython with pip, and build c object file with below command.
# python cython_setup.py build_ext --inplace


import cupy as cp
import numpy as np
from itertools import product, islice


CP_NP_CRITERIA = 50


def get_gpu_mem_size():
    gpu_count = cp.cuda.runtime.getDeviceCount()
    free_mem, total_mem = cp.cuda.Device(0).mem_info
    free_mem /= 1000 * 1000 * 1000
    return int(free_mem)


def likelihood(crop_imgs, gauss_grid, bg_squared_sums, bg_means, window_size1, window_size2):
    nb_imgs = crop_imgs.shape[0]
    if nb_imgs < CP_NP_CRITERIA:
        crop_imgs = cp.asarray(crop_imgs)
    gauss_grid = cp.asarray(gauss_grid)
    bg_squared_sums = cp.asarray(bg_squared_sums)
    bg_means = cp.asarray(bg_means)
    surface_window = window_size1 * window_size2
    g_mean = cp.mean(gauss_grid)
    g_bar = (gauss_grid - g_mean).reshape([window_size1 * window_size2, 1])
    g_squared_sum = cp.sum(g_bar ** 2)
    i_hat = (crop_imgs - bg_means.reshape(crop_imgs.shape[0], 1, 1))
    i_local_mins = cp.min(i_hat, axis=(1, 2))
    i_hat -= cp.maximum(cp.zeros_like(i_local_mins), i_local_mins).reshape(-1, 1, 1)
    i_hat = cp.matmul(i_hat, g_bar) / g_squared_sum
    i_hat = cp.maximum(cp.zeros(i_hat.shape), i_hat)
    L = ((surface_window / 2.) * cp.log(1 - (i_hat ** 2 * g_squared_sum).T /
                                        (bg_squared_sums - (surface_window * bg_means)))).T
    L = cp.asnumpy(L.reshape(crop_imgs.shape[0], crop_imgs.shape[1], 1).astype(cp.float32))
    return L


def background(imgs, window_sizes, alpha):
    imgs = cp.asarray(imgs)
    imgs = imgs / cp.max(imgs, axis= (1,2)).reshape(-1, 1, 1)
    bins = 0.01
    nb_imgs = imgs.shape[0]
    img_flat_length = imgs.shape[1] * imgs.shape[2]
    bgs = {}
    bg_means = []
    bg_stds = []
    bg_intensities = (imgs.reshape(nb_imgs, img_flat_length) * 100).astype(cp.uint8) / 100

    for idx in range(nb_imgs):
        if cp.sum(bg_intensities[idx]) <= 0:
            dummy_intensities = 0.1
            bg_intensities[idx][0] = dummy_intensities

    post_mask_args = cp.array([cp.arange(img_flat_length) for _ in range(nb_imgs)], dtype=cp.int64)
    mask_sums_modes = cp.zeros(nb_imgs)
    mask_stds = cp.empty(nb_imgs)
    for repeat in range(3):
        for i in range(nb_imgs):
            if len(post_mask_args[i]) <= 0:
                print('Errors on images, please check images again whether it contains an empty black-image. If not, contact the author.')
                continue
            it_hist, bin_width = cp.histogram(bg_intensities[i, post_mask_args[i]], bins=cp.arange(0, cp.max(bg_intensities[i, post_mask_args[i]]) + bins, bins))
            if len(it_hist) < 1:
                print('Errors on images, please check images again whether it contains an empty black-image. If not, contact the author.')
                continue
            mask_sums_modes[i] = (cp.argmax(it_hist) * bins + (bins / 2))
        if repeat==0:
            mask_stds = cp.std(cp.take(bg_intensities, post_mask_args), axis=1)
        else:
            for i in range(nb_imgs):
                mask_stds[i] = cp.std(cp.take(bg_intensities[i], post_mask_args[i]))
        post_mask_args = []
        for i in range(nb_imgs):
            post_mask_args.append(cp.argwhere((bg_intensities[i] > float(mask_sums_modes[i] - 3. * mask_stds[i])) & (bg_intensities[i] < float(mask_sums_modes[i] + 3. * mask_stds[i]))).flatten())
    for i in range(nb_imgs):
        bg_means.append(cp.mean(cp.take(bg_intensities[i], post_mask_args[i])))
        bg_stds.append(cp.std(cp.take(bg_intensities[i], post_mask_args[i])))
    bg_means = cp.array(bg_means, dtype=cp.float32)
    bg_stds = cp.array(bg_stds, dtype=cp.float32)
    for window_size in window_sizes:
        bg = cp.ones((bg_intensities.shape[0], window_size[0] * window_size[1]), dtype=cp.float32)
        bg *= bg_means.reshape(-1, 1)
        bgs[window_size[0]] = cp.asnumpy(bg)

    thresholds = cp.asnumpy(1/(bg_means**2 / bg_stds**2)) * 2.0
    thresholds = np.maximum(thresholds, np.ones_like(thresholds) * 1.0)
    for th_i in range(len(thresholds)):
        if np.isnan(thresholds[th_i]):
            thresholds[th_i] = 1.0
    return bgs, thresholds * alpha


def image_cropping(extended_imgs, extend, window_size0, window_size1, shift):
    nb_imgs = extended_imgs.shape[0]
    row_size = extended_imgs.shape[1]
    col_size = extended_imgs.shape[2]
    start_row = int(extend/2 - (window_size1-1)/2)
    end_row = row_size - window_size1 - start_row + 1
    start_col = int(extend/2 - (window_size0-1)/2)
    end_col = col_size - window_size0 - start_col + 1
    row_col_comb = list(product(range(start_row, end_row, shift), range(start_col, end_col, shift)))
    if nb_imgs >= CP_NP_CRITERIA:
        extended_imgs = cp.asarray(extended_imgs)
        cropped_imgs = cp.empty([nb_imgs, len(row_col_comb), window_size0, window_size1], dtype=cp.float32)
    else:
        cropped_imgs = np.empty([nb_imgs, len(row_col_comb), window_size0, window_size1], dtype=np.float32)
    index = 0
    for r, c in row_col_comb:
        r = int(r)
        c = int(c)
        cropped_imgs[:, index] = extended_imgs[:, r:r + window_size1, c:c + window_size0]
        index += 1
    return cropped_imgs.reshape(nb_imgs, -1, window_size0 * window_size1), type(cropped_imgs)==cp.ndarray


def chunk(arr_range, arr_size):
    arr_range = iter(arr_range)
    return iter(lambda: tuple(islice(arr_range, arr_size)), ())



# Modified by Claude (claude-opus-4-6, Anthropic AI) — lazy model download on first use
import os
import re
import numpy as np
from tensorflow.keras.models import load_model # type: ignore
from FreeTrace.model_downloader import ensure_models


class RegModel:
    def __init__(self, reg_model_nums:list):
        self.reg_model_nums = reg_model_nums
        self.alpha_models = None
        self.k_model = None
        self.crits = reg_model_nums.copy()
        self.crits.append(8192)
        self.load_models(self.reg_model_nums)

    def load_models(self, model_nums):
        ensure_models()
        freetrace_path = ''
        freetrace_path += 'FreeTrace'.join(re.split(r'FreeTrace', __file__)[:-1]) + 'FreeTrace'
        self.alpha_models = {n: load_model(f'{freetrace_path}/models/reg_model_{n}.keras') for n in model_nums}
        self.k_model = load_model(f'{freetrace_path}/models/reg_k_model.keras')

        for n in model_nums:
            self.alpha_models[n].compile()
        self.k_model.compile()

    def call(self, inputs):
        x = inputs[0]
        y = inputs[1]
        input_signals = []
        model_num = self.model_selection(x.shape[0])
        re_couped_x, re_couped_y = self.recoupe_trajectory(x, y, model_num)
        for r_x, r_y in zip(re_couped_x, re_couped_y):
            input_signal1, input_signal2 = self.cvt_2_signal(r_x, r_y)
            input_signals.append(input_signal1)
            input_signals.append(input_signal2)
        input_signals = np.reshape(input_signals, [-1, model_num, 1, 3])
        return input_signals, model_num

    def alpha_predict(self, inputs):
        if len(inputs) == 2:
            if len(inputs[0]) < self.reg_model_nums[0]:
                return 1.0
            alpha_signal, model_num = self.call(inputs)
            pred_alphas = self.alpha_models[model_num].predict_on_batch(alpha_signal)
            if pred_alphas.shape[0] > 4:
                pred_alpha = np.mean(np.quantile(pred_alphas, q=[0.25, 0.75], method='normal_unbiased'))
            else:
                pred_alpha = np.mean(pred_alphas)
            return pred_alpha
        elif len(inputs) == 3:
            if len(inputs[0]) < self.reg_model_nums[0]:
                return 1.0
            alpha_preds = []
            for inputs_ in inputs:
                alpha_signal, model_num = self.call(inputs_)
                pred_alphas = self.alpha_models[model_num].predict_on_batch(alpha_signal)
                if pred_alphas.shape[0] > 4:
                    pred_alpha = np.mean(np.quantile(pred_alphas, q=[0.25, 0.75], method='normal_unbiased'))
                else:
                    pred_alpha = np.mean(pred_alphas)
                alpha_preds.append(pred_alpha)
            return alpha_preds

    def k_predict(self, inputs):
        log_disps = []
        for input_ in inputs:
            log_disps.append(self.log_displacements(input_[0], input_[1]))
        k_preds = self.k_model.predict_on_batch(np.array(log_disps))
        for k_idx in range(len(k_preds)):
            if np.isnan(k_preds[k_idx]):
                k_preds[k_idx] = 1.0
        return k_preds.flatten()

    def model_selection(self, length):
        index = 0
        while True:
            if self.crits[index] <= length < self.crits[index+1]:
                return self.reg_model_nums[index]
            index += 1
            if index >= len(self.crits):
                return self.reg_model_nums[-2]

    def log_displacements(self, xs, ys):
        disps = self.displacement(xs, ys)
        if xs.shape[0] < 10:
            return np.log10(np.mean(disps))
        else:
            return np.log10(np.mean(np.quantile(disps, q=[0.25, 0.75], method='normal_unbiased')))

    def recoupe_trajectory(self, x, y, model_num, jump=1):
        couped_x = []
        couped_y = []
        for i in range(0, x.shape[0], jump):
          tmp1 = x[i: i + model_num]
          tmp2 = y[i: i + model_num]
          if tmp1.shape[0] == model_num:
              couped_x.append(tmp1)
              couped_y.append(tmp2)
        return couped_x, couped_y

    def cvt_2_signal(self, x, y):
        rad_list, xs_raw, ys_raw = self.make_alpha_inputs(x, y)
        x = x / (np.std(x))
        x = np.cumsum(self.abs_subtraction(x)) / x.shape[0]
        y = y / (np.std(y))
        y = np.cumsum(self.abs_subtraction(y)) / y.shape[0]
        return np.transpose(np.stack((x, rad_list, xs_raw))), np.transpose(np.stack((y, rad_list, ys_raw)))

    def abs_subtraction(self, xs):
        assert xs.ndim == 1
        uncum_list = [0.]
        for i in range(1, xs.shape[0]):
            uncum_list.append(abs(xs[i] - xs[i-1]))
        return uncum_list

    def make_alpha_inputs(self, xs, ys):
        assert xs.ndim == 1 and ys.ndim == 1
        rad_list = self.radius(xs, ys)
        disp_list = self.displacement(xs, ys)
        return (rad_list / np.mean(disp_list) / xs.shape[0],
                ((xs - xs[0]) / np.mean(disp_list) / xs.shape[0]),
                ((ys - ys[0]) / np.mean(disp_list) / ys.shape[0]))

    def radius(self, xs, ys):
        rad_list = [0.]
        for i in range(1, xs.shape[0]):
            rad_list.append(np.sqrt((xs[i] - xs[0])**2 + (ys[i] - ys[0])**2))
        return rad_list

    def displacement(self, xs, ys):
        disp_list = []
        for i in range(1, xs.shape[0]):
            disp_list.append(np.sqrt((xs[i] - xs[i-1])**2 + (ys[i] - ys[i-1])**2))
        return disp_list

    def get_reg_model_nums(self):
        return self.reg_model_nums
